@@ -74,6 +74,7 @@ Memoization
 .. autofunction:: memoize_in
 .. autofunction:: keyed_memoize_on_first_arg
 .. autofunction:: keyed_memoize_method
+.. autofunction:: keyed_memoize_in
 
 Argmin/max
 ----------
@@ -693,11 +694,15 @@ def memoize_on_first_arg(function, cache_dict_name=None):
         try:
             return getattr(obj, cache_dict_name)[key]
         except AttributeError:
-            result = function(obj, *args, **kwargs)
+            attribute_error = True
+        except KeyError:
+            attribute_error = False
+
+        result = function(obj, *args, **kwargs)
+        if attribute_error:
             object.__setattr__(obj, cache_dict_name, {key: result})
             return result
-        except KeyError:
-            result = function(obj, *args, **kwargs)
+        else:
             getattr(obj, cache_dict_name)[key] = result
             return result
 
@@ -717,7 +722,7 @@ def memoize_method(method: F) -> F:
     .. versionchanged:: 2021.2
 
         Can memoize methods on classes that do not allow setting attributes
-        (e.g. by overwritting ``__setattr__``).
+        (e.g. by overwritting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
     """
 
     return memoize_on_first_arg(method,
@@ -732,6 +737,8 @@ class keyed_memoize_on_first_arg:  # noqa: N801
 
     :arg key: A function receiving the same arguments as the decorated function
         which computes and returns the cache key.
+    :arg cache_dict_name: The name of the `dict` attribute in the instance
+        used to hold the cache.
 
     .. versionadded :: 2020.3
     """
@@ -775,7 +782,10 @@ class keyed_memoize_on_first_arg:  # noqa: N801
 
 
 class keyed_memoize_method(keyed_memoize_on_first_arg):  # noqa: N801
-    """Supports cache deletion via ``method_name.clear_cache(self)``.
+    """Like :class:`memoize_method`, but additionally uses a function *key* to
+    compute the key under which the function result is stored.
+
+    Supports cache deletion via ``method_name.clear_cache(self)``.
 
     :arg key: A function receiving the same arguments as the decorated function
         which computes and returns the cache key.
@@ -785,7 +795,7 @@ class keyed_memoize_method(keyed_memoize_on_first_arg):  # noqa: N801
     .. versionchanged:: 2021.2
 
         Can memoize methods on classes that do not allow setting attributes
-        (e.g. by overwritting ``__setattr__``).
+        (e.g. by overwritting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
     """
     def _default_cache_dict_name(self, function):
         return intern(f"_memoize_dic_{function.__name__}")
@@ -859,55 +869,28 @@ def memoize_method_with_uncached(uncached_args=None, uncached_kwargs=None):
     return parametrized_decorator
 
 
-def memoize_method_nested(inner):
-    """Adds a cache to a function nested inside a method. The cache is attached
-    to *memoize_cache_context* (if it exists) or *self* in the outer (method)
-    namespace.
-
-    Requires Python 2.5 or newer.
-    """
-
-    from warnings import warn
-    warn("memoize_method_nested is deprecated and will go away in 2021. "
-            "Use @memoize_in(self, 'identifier') instead", DeprecationWarning,
-            stacklevel=2)
-
-    cache_dict_name = intern("_memoize_inner_dic_%s_%s_%d"
-            % (inner.__name__, inner.__code__.co_filename,
-                inner.__code__.co_firstlineno))
-
-    from inspect import currentframe
-    outer_frame = currentframe().f_back
-    cache_context = outer_frame.f_locals.get("memoize_cache_context")
-    if cache_context is None:
-        cache_context = outer_frame.f_locals.get("self")
-
-    try:
-        cache_dict = getattr(cache_context, cache_dict_name)
-    except AttributeError:
-        cache_dict = {}
-        setattr(cache_context, cache_dict_name, cache_dict)
-
-    @wraps(inner)
-    def new_inner(*args):
-        try:
-            return cache_dict[args]
-        except KeyError:
-            result = inner(*args)
-            cache_dict[args] = result
-            return result
-
-    return new_inner
-
-
 class memoize_in:  # noqa
-    """Adds a cache to a function nested inside a method. The cache is attached
-    to *container*.
+    """Adds a cache to the function it decorates. The cache is attached
+    to *container* and must be uniquely specified by *identifier* (i.e.
+    all functions using the same *container* and *identifier* will be using
+    the same cache). The decorated function may only receive positional
+    arguments.
+
+    .. note::
+
+        This function works well on nested functions, which
+        do not have stable global identifiers.
 
     .. versionchanged :: 2020.3
 
         *identifier* no longer needs to be a :class:`str`,
         but it needs to be hashable.
+
+    .. versionchanged:: 2021.2.1
+
+        Can now use instances of classes as *container* that do not allow
+        setting attributes (e.g. by overwritting ``__setattr__``),
+        e.g. frozen :mod:`dataclasses`.
     """
 
     def __init__(self, container, identifier):
@@ -915,7 +898,8 @@ class memoize_in:  # noqa
             memoize_in_dict = container._pytools_memoize_in_dict
         except AttributeError:
             memoize_in_dict = {}
-            container._pytools_memoize_in_dict = memoize_in_dict
+            object.__setattr__(container, "_pytools_memoize_in_dict",
+                    memoize_in_dict)
 
         self.cache_dict = memoize_in_dict.setdefault(identifier, {})
 
@@ -931,6 +915,41 @@ class memoize_in:  # noqa
 
         return new_inner
 
+
+class keyed_memoize_in:  # noqa
+    """Like :class:`memoize_in`, but additionally uses a function *key* to
+    compute the key under which the function result is memoized.
+
+    :arg key: A function receiving the same arguments as the decorated function
+        which computes and returns the cache key.
+
+    .. versionadded :: 2021.2.1
+    """
+
+    def __init__(self, container, identifier, key):
+        try:
+            memoize_in_dict = container._pytools_keyed_memoize_in_dict
+        except AttributeError:
+            memoize_in_dict = {}
+            object.__setattr__(container, "_pytools_keyed_memoize_in_dict",
+                    memoize_in_dict)
+
+        self.cache_dict = memoize_in_dict.setdefault(identifier, {})
+        self.key = key
+
+    def __call__(self, inner):
+        @wraps(inner)
+        def new_inner(*args):
+            key = self.key(*args)
+            try:
+                return self.cache_dict[key]
+            except KeyError:
+                result = inner(*args)
+                self.cache_dict[key] = result
+                return result
+
+        return new_inner
+
 # }}}
 
 
@@ -2419,7 +2438,11 @@ class ProcessLogger:  # pylint: disable=too-many-instance-attributes
             # Can happen, e.g., if pudb is controlling the console.
             use_late_start_logging = False
         else:
-            use_late_start_logging = sys.stdin.isatty()
+            if hasattr(sys.stdin, "closed") and not sys.stdin.closed:
+                # can query stdin.isatty() only if stdin's open
+                use_late_start_logging = sys.stdin.isatty()
+            else:
+                use_late_start_logging = False
 
         import os
         if os.environ.get("PYTOOLS_LOG_NO_THREADS", ""):
